import re

import requests


def _clean_links(links):
  deduped = set()
  for domain in links:
    lower = (domain or '').lower()
    split = lower.split('<br>')
    for full_domain in split:
      deduped.add(full_domain.strip())
  return list(deduped)


def search_crtsh(self, target):
  print("Searching crt.sh")
  headers = {'authority':                 'crt.sh',
             'cache-control':             'max-age=0',
             'upgrade-insecure-requests': '1',
             'user-agent':                'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.28 Safari/537.36',
             'sec-metadata':              'cause=forced, destination=document, site=cross-site',
             'sec-origin-policy':         '0', 'dnt': '1',
             'accept':                    'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
             'accept-encoding':           'gzip, deflate, br',
             'accept-language':           'en-US,en;q=0.9,it;q=0.8,la;q=0.7', }

  params = (('q', '%.' + target),)

  try:
    res = requests.get('https://crt.sh/', headers=headers, params=params)
    scraped = res.text
    subdomain_finder = re.compile('<TD>(.*\.' + target + ')</TD>')
    links = subdomain_finder.findall(scraped)
    parsed_links = _clean_links(links)

    for domain in parsed_links:
      if domain.strip() not in self.domains and domain.endswith("." + target):
        self.domains.append(domain.strip())
        if self.options["--verbose"]:
          print("Crt.sh Found Domain:", domain.strip())
  except Exception as e:
    self.handle_exception(e, "Error searching crt.sh")



import os

from nmap import nmap

from anubis.utils.ColorPrint import ColorPrint


def dnssecc_subdomain_enum(self, target):
  # Must run as root
  if os.getuid() == 0:
    print("Starting DNSSEC Enum")
    nm = nmap.PortScanner()
    arguments = '-sSU -p 53 --script dns-nsec-enum --script-args dns-nsec-enum.domains=' + target

    nm.scan(hosts=self.ip, arguments=arguments)
    for host in nm.all_hosts():
      try:
        print(nm[host]['udp'][53]['script']['dns-nsec-enum'])
      except:
        pass
  else:
    ColorPrint.red(
      "To run a DNSSEC subdomain enumeration, Anubis must be run as root")



import sys


class SignalHandler:
  """
  The object that will handle signals and stop the worker threads.
  """

  #: The stop event that's shared by this handler and threads.
  stopper = None

  #: The pool of worker threads
  workers = None

  def __init__(self, stopper, workers):
    self.stopper = stopper
    self.workers = workers

  def __call__(self, signum, frame):
    """
    This will be called by the python signal module

    https://docs.python.org/3/library/signal.html#signal.signal
    """
    self.stopper.set()

    for worker in self.workers:
      worker.join()

    sys.__stdout__.write("Quitting...")
    sys.__stdout__.flush()
    sys.exit(0)



import shutil
from json import dumps

import nmap


# Performs an nmap scan of a target, and outputs interesting services/ssl information
def scan_host(self):
  if shutil.which("nmap") is None:
    print("Nmap must be installed to run --with-nmap!")
    return

  print("Starting Nmap Scan")
  nm = nmap.PortScanner()
  arguments = '-nPn -sV -sC'
  if self.options["--overwrite-nmap-scan"]:
    arguments = self.options["--overwrite-nmap-scan"]
  nm.scan(hosts=self.ip, arguments=arguments)
  # Could be extended to scan multiple hosts in the future
  for host in nm.all_hosts():
    if self.options["--verbose"]:
      print(dumps(nm[host], indent=2, sort_keys=True))
    print('Host : %s (%s)' % (host, nm[host].hostname()))
    for proto in nm[host].all_protocols():
      print('----------')
    print('Protocol: %s' % proto)
    lport = nm[host][proto].keys()
    for port in lport:
      print('port: %s\tstate: %s' % (port, nm[host][proto][port]['state']))
      try:
        if nm[host][proto][port]['product']:
          print('\tservice: %s' % nm[host][proto][port]['product'], end='')
          if nm[host][proto][port]['version']:
            print('\tversion: %s' % nm[host][proto][port]['version'])
          else:
            print('')
      except Exception as e:
        self.handle_exception(e)
      try:
        fix_newline = nm[host][proto][port]['script']['ssl-cert'].split('\n')
        print('\tCertificate:')
        for i in range(len(fix_newline)):
          print('\t', fix_newline[i])
          if fix_newline[i].startswith("Subject Alternative Name: "):
            content = fix_newline[i].replace("Subject Alternative Name: ", '')
            content = content.replace("DNS:", '')
            new_domains = content.split(",")
            for domain in new_domains:
              domain = domain.strip()
              if domain not in self.domains:
                self.domains.append(domain)
                if self.options["--verbose"]:
                  print("Nmap Found Domain:", domain.strip())
      except Exception as e:
        self.handle_exception(e)



from collections import defaultdict
import socket
import ssl
from socket import gaierror


def search_subject_alt_name(self, target):
  print("Searching for Subject Alt Names")
  try:
    context = ssl.create_default_context()

    # Do connectivity testing to ensure SSLyze is able to connect
    try:
      with socket.create_connection((target, 443)) as sock:
        with context.wrap_socket(sock, server_hostname=target) as ssock:
          # https://docs.python.org/3/library/ssl.html#ssl.SSLSocket.getpeercert
          cert = ssock.getpeercert()

      subjectAltName = defaultdict(set)
      for type_, san in cert['subjectAltName']:
        subjectAltName[type_].add(san)

      dns_domains = list(subjectAltName['DNS'])
      for domain in dns_domains:
        if domain:
          self.domains.append(domain.strip())
    except gaierror as e:
      # Could not connect to the server; abort
      print(f"Error connecting to {target}: {e}")
      return

  except Exception as e:
    self.handle_exception(e)



"""
Usage:
  anubis (-t TARGET | -f FILE) [-o FILENAME]  [-abinoprsSv] [-w SCAN] [-q NUM]
  anubis -h
  anubis --version
  
Options:
  -h --help                       show this help message and exit
  -t --target                     set target (comma separated, no spaces, if multiple)
  -f --file                       set target (reads from file, one domain per line)
  -n --with-nmap                  perform an nmap service/script scan
  -o --output                     save to filename
  -i --additional-info            show additional information about the host from Shodan (requires API key)
  -p --ip                         outputs the resolved IPs for each subdomain, and a full list of unique ips
  -a --send-to-anubis-db          send results to Anubis-DB
  -r --recursive                  recursively search over all subdomains
  -s --ssl                        run an ssl scan and output cipher + chain info
  -S --silent                     only out put subdomains, one per line
  -w --overwrite-nmap-scan SCAN   overwrite default nmap scan (default -nPn -sV -sC)
  -v --verbose                    print debug info and full request output
  -q --queue-workers NUM          override number of queue workers (default: 10, max: 100)
  --version                       show version and exit

Help:
  For help using this tool, please open an issue on the Github repository:
  https://github.com/jonluca/anubis
"""

import sys
import time
from functools import reduce

from docopt import docopt

# noinspection PyPep8Naming
from . import __version__ as VERSION


# Overload stdout to save output and change colors on filewrite
class StdOutHook:
  lines = []
  filename = ""

  def __init__(self, filename, silent, output):
    self.filename = filename
    self.silent = silent
    self.output = output

  def write(self, text, override=False, **kwargs):
    if not self.silent or override:
      sys.__stdout__.write(text)
    self.lines.append(text)

  def write_out(self):
    if self.output:
      with open(self.filename, "w") as file:
        for line in self.lines:
          # remove stdout colors
          line = line.replace('\033[91m', '')
          line = line.replace('\033[92m', '')
          line = line.replace('\033[93m', '')
          line = line.replace('\033[94m', '')
          line = line.replace('\033[95m', '')
          line = line.replace('\033[0m', '')
          file.write(line)

  def flush(self):
    # python3 compatability, does nothing
    pass


# credit to https://stackoverflow.com/questions/1557571/how-do-i-get-time-of-a-python-programs-execution
def seconds_to_str(t):
  return "%d:%02d:%02d.%03d" % reduce(lambda ll, b: divmod(ll[0], b) + ll[1:],
                                      [(t * 1000,), 1000, 60, 60])


def main():
  try:
    if sys.version_info < (3, 0):
      sys.stdout.write("Sorry, requires Python 3.x, not Python 2.x\n")
      sys.exit(1)

    start_time = time.time()

    import anubis.commands

    options = docopt(__doc__, version=VERSION)

    if options["--output"] or options['--silent']:
      sys.stdout = StdOutHook(options["FILENAME"], options['--silent'],
                              options['--output'])

    if options["--queue-workers"]:
      if not options["--recursive"]:
        print("Queue workers override only works with recursive option!")
        sys.exit(1)
      if int(options["--queue-workers"]) > 100:
        print("Max queue worker override is 100!")
        sys.exit(1)
      if int(options["--queue-workers"]) < 1:
        print("Queue workers can't be negative!")
        sys.exit(1)

    if not options["--target"] and not options['--file']:
      print(
        "Target required! Run with -h for usage instructions. Either -t target.host or -f file.txt required")
      return

    if options["--target"] and options['--file']:
      print(
        "Please only supply one target method - either read by file with -f or as an argument to -t, not both.")
      return

    print("""
          d8888                   888      d8b
         d88888                   888      Y8P
        d88P888                   888
       d88P 888 88888b.  888  888 88888b.  888 .d8888b
      d88P  888 888 "88b 888  888 888 "88b 888 88K
     d88P   888 888  888 888  888 888  888 888 "Y8888b.
    d8888888888 888  888 Y88b 888 888 d88P 888      X88
   d88P     888 888  888  "Y88888 88888P"  888  88888P'
    """)

    command = anubis.commands.Target(options)
    command.run()
    print("Subdomain search took %s" % seconds_to_str(time.time() - start_time))
    if options["--output"]:
      sys.stdout.write_out()
  except KeyboardInterrupt:
    print("\nQuitting...")
    sys.exit(0)



"""Tests for our main anubis CLI module."""

import os
import signal
import sys
from io import StringIO
from subprocess import PIPE, Popen as popen
from threading import Timer
from time import sleep
from unittest import TestCase

from anubis.commands.target import Target
from anubis.scanners.anubis_db import search_anubisdb, send_to_anubisdb
from anubis.scanners.crt import search_crtsh
from anubis.scanners.dnsdumpster import search_dnsdumpster
from anubis.scanners.dnssec import dnssecc_subdomain_enum
from anubis.scanners.hackertarget import subdomain_hackertarget
from anubis.scanners.netcraft import search_netcraft
from anubis.scanners.pkey import search_pkey
from anubis.scanners.recursive import recursive_search
from anubis.scanners.shodan import search_shodan
from anubis.scanners.zonetransfer import dns_zonetransfer
from anubis.utils.ColorPrint import ColorPrint


class TestScanners(TestCase):
  domains = list()
  options = {"--verbose": True}

  def handle_exception(self, e, message=""):
    print(e)
    if message:
      print(message)

  def setUp(self):
    # catch stdout
    self.held, sys.stdout = sys.stdout, StringIO()
    self.stdout = sys.stdout
    # reset domains
    self.domains = list()

  def tearDown(self):
    self.held.write(sys.stdout.getvalue())

  def test_anubis_db(self):
    search_anubisdb(self, "example.com")
    self.assertIn("www.example.com", self.domains)

  def test_send_to_anubis(self):
    self.domains.append("www.example.com")
    send_to_anubisdb(self,
                     ["example.com"])  # Send to anubis db takes in an array
    self.assertTrue("Error" not in sys.stdout.getvalue())

    send_to_anubisdb(self, ["example.com", "www.example.com"])
    self.assertTrue("multiple" in sys.stdout.getvalue())

    send_to_anubisdb(self, ["example"])
    self.assertTrue("Error" in sys.stdout.getvalue())

  def test_crt(self):
    search_crtsh(self, "jonlu.ca")

    if 'TRAVIS' in os.environ:
      self.assertTrue(True) # crt.sh times out on Travis
      return
    self.assertIn("secure.jonlu.ca", self.domains)

  def test_zonetransfer(self):
    dns_zonetransfer(self, "jonlu.ca")
    self.assertTrue("Error" not in sys.stdout.getvalue())
    dns_zonetransfer(self, "$FALSE$")
    self.assertTrue("Error" in sys.stdout.getvalue())

  def test_dnsdumpster(self):
    search_dnsdumpster(self, "example.com")
    self.assertIn("www.example.com", self.domains)

  def test_dnssec(self):
    if os.getuid() == 0:
      dnssecc_subdomain_enum(self, "google.com")
      self.assertTrue("google" in sys.stdout.getvalue())
    else:
      print("To run DNSSEC test, run as root")

  def test_hackertarget(self):
    subdomain_hackertarget(self, "example.com")
    self.assertIn("www.example.com", self.domains)

  def test_netcraft(self):
    search_netcraft(self, "example.com")
    self.assertTrue(True) # patch after netcraft no long returns valid results, 1/6/19
    # self.assertIn("http://www.example.com", self.domains)

  # As of 1/18/18, Pkey currently constantly times out
  def test_pkey(self):
    search_pkey(self, "google.com")
    # self.assertIn("google.com", self.domains)
    self.assertTrue(True)

  def test_shodan(self):
    self.ip = "138.197.125.24"
    self.options = {}
    self.options["TARGET"] = ["jonlu.ca"]
    self.options["--verbose"] = True
    search_shodan(self)
    self.assertIn("ISP", sys.stdout.getvalue())

  def test_recursive(self):
    self.options = {}
    # Set target to domain we know only has 1 subdomain, for speeds sake
    self.options["TARGET"] = ["neverssl.com"]
    self.domains.append("neverssl.com")
    self.options["--verbose"] = True
    self.options["--queue-workers"] = False
    recursive_search(self)
    self.domains = self.clean_domains(self.domains)
    self.assertIn("www.neverssl.com", self.domains)

  # Pass through function for recursive search
  def clean_domains(self, domains):
    return Target.clean_domains(domains)

  def test_sigints(self):
    # Declare function to send sigint, after timer

    proc1 = popen(['anubis', '-tr', 'neverssl.com'], stdout=PIPE)

    # Function to send sigint to our processes, make sure that it outputss "Quitting" then ends
    def send_siginit():
      popen.send_signal(proc1, signal.SIGINT)
      self.assertTrue("Quitting" in sys.stdout.getvalue())

    t = Timer(3.0, send_siginit)
    t.start()
    sleep(5)

  def test_exception(self):
    self.options["--verbose"] = True
    try:
      raise Exception("Oh No")
    except Exception as e:
      Target.handle_exception(self, e, "Test Exception")
      self.assertTrue("Test" in sys.stdout.getvalue())


class TestColorPrint(TestCase):

  def setUp(self):
    # catch stdout
    self.held, sys.stdout = sys.stdout, StringIO()
    # reset domains
    self.domains = list()

  def tearDown(self):
    # Kind of hacky as it'll dump all of stdout at the end, but better than losing it entirely
    self.held.write(sys.stdout.getvalue())

  def test_color_print(self):
    ColorPrint.red("red")
    self.assertIn("91m", sys.stdout.getvalue())
    ColorPrint.green("green")
    self.assertIn("92m", sys.stdout.getvalue())
    ColorPrint.light_purple("light_purple")
    self.assertIn("94m", sys.stdout.getvalue())
    ColorPrint.purple("purple")
    self.assertIn("95m", sys.stdout.getvalue())
    ColorPrint.yellow("yellow")
    self.assertIn("93m", sys.stdout.getvalue())



class ColorPrint:
  RED = '\033[91m'
  GREEN = '\033[92m'
  YELLOW = '\033[93m'
  LIGHT_PURPLE = '\033[94m'
  PURPLE = '\033[95m'
  END = '\033[0m'

  @classmethod
  def red(self, s, **kwargs):
    print(self.RED + s + self.END, **kwargs)

  @classmethod
  def green(self, s, **kwargs):
    print(self.GREEN + s + self.END, **kwargs)

  @classmethod
  def yellow(self, s, **kwargs):
    print(self.YELLOW + s + self.END, **kwargs)

  @classmethod
  def light_purple(self, s, **kwargs):
    print(self.LIGHT_PURPLE + s + self.END, **kwargs)

  @classmethod
  def purple(self, s, **kwargs):
    print(self.PURPLE + s + self.END, **kwargs)



import re

import requests


def search_dnsdumpster(self, target):
  print("Searching DNSDumpster")
  headers = {'Pragma':          'no-cache', 'Origin': 'https://dnsdumpster.com',
             'Accept-Encoding': 'gzip, deflate, br',
             'Accept-Language': 'en-US,en;q=0.9,it;q=0.8',
             'User-Agent':      'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36',
             'Content-Type':    'application/x-www-form-urlencoded',
             'Accept':          'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
             'Cache-Control':   'no-cache',
             'Referer':         'https://dnsdumpster.com/',
             'Connection':      'keep-alive', 'DNT': '1', }

  get_csrf_res = requests.get('https://dnsdumpster.com', headers=headers)

  try:
    csrf_token = get_csrf_res.headers['Set-Cookie']
    csrf_token = csrf_token[10:]
    csrf_token = csrf_token.split(";")[0]
  except Exception as e:
    self.handle_exception(e, "Retrieving CSRF Token for DNSDumpster failed")
    return

  cookies = {'csrftoken': csrf_token, }

  data = [('csrfmiddlewaretoken', csrf_token), ('targetip', target)]

  res = requests.post('https://dnsdumpster.com/', headers=headers,
                      cookies=cookies, data=data)
  try:
    scraped = res.text
    subdomain_finder = re.compile('\">(.*\.' + target + ')<br>')
    links = subdomain_finder.findall(scraped)
    for domain in links:
      if domain.strip() not in self.domains and domain.endswith("." + target):
        self.domains.append(domain.strip())
        if self.options["--verbose"]:
          print("DNSDumpster Found Domain:", domain.strip())
  except Exception as e:
    self.handle_exception(e, "Error searching DNS Dumpster")
    pass
